
// Add long vectors using CUDA
// Uses a 2D grid layout; if you have insanely huge vectors, you'll need to make your own 3D layout ;-)
//
// nvcc -arch=sm_35 --compiler-options -Wall vectoradd_2d.cu -o vectoradd_2d
//
// GZ, Feb 2013


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#ifdef _WIN32
#define srand48(expr) srand(expr)
#define drand48() double(rand())/double(RAND_MAX)
#endif

// Add two vectors 'a' and 'b' of length 'n', put the result in c.
// Prerequisite: 'c'c must hold at least 'n' floats!

__global__
void addVectors( const float *a, const float *b, float *c, unsigned int n )
{
	unsigned int tid_x = blockDim.x * blockIdx.x + threadIdx.x;		// tid = thread ID within block
	unsigned int tid_y = blockDim.y * blockIdx.y + threadIdx.y;
	unsigned int i = tid_y * (blockDim.x * gridDim.x) + tid_x;		// blockDim.x * gridDim.x = #threads per row

	if ( i < n )
		c[i] = a[i] + b[i];
}



// General CUDA error handler
// Prints error message, if something went wrong
void handleCUDAerror( int line )
{
	cudaError_t err = cudaGetLastError();
	if ( err == cudaSuccess )
		return;

	printf("CUDA error message: %s\n"
		   "line %d\n", cudaGetErrorString(err), line );

	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
    printf("CUDA device %d: \"%s\" with compute capability %d.%d\n",
           devID, props.name, props.major, props.minor);
}


// Check whether our 2D grid layout fits with the current GPU's maximum properties
// Warning: considers only X and Y directions !
void checkMaxProperties( dim3 blocks, dim3 threads )
{
	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
	handleCUDAerror(__LINE__);

	if ( threads.x * threads.y > unsigned(props.maxThreadsPerBlock) )
	{
		fprintf(stderr, "Too many threads per block! (%d x %d > %d)\n",
				threads.x, threads.y, props.maxThreadsPerBlock );
		exit(1);
	}
	if ( threads.x > unsigned(props.maxThreadsDim[0]) )
	{
		fprintf(stderr, "Too many threads per block along X direction! (%d > %d)\n",
				threads.x, props.maxThreadsDim[0] );
		exit(1);
	}
	if ( threads.y > unsigned(props.maxThreadsDim[1]) )
	{
		fprintf(stderr, "Too many threads per block along Y direction! (%d > %d)\n",
				threads.y, props.maxThreadsDim[1] );
		exit(1);
	}

	if ( blocks.x > unsigned(props.maxGridSize[0]) )
	{
		fprintf(stderr, "Too many blocks in grid along X direction! (%d > %d)\n",
				blocks.x, props.maxGridSize[0] );
		exit(1);
	}
	if ( blocks.y > unsigned(props.maxGridSize[1]) )
	{
		fprintf(stderr, "Too many blocks in grid along Y direction! (%d > %d)\n",
				blocks.y, props.maxGridSize[1] );
		exit(1);
	}

	// all's well ...
}



int main( int argc, char **argv )
{
	unsigned int vec_len = 10000;									// just a default
	if ( argc > 1 )
		// interpret first argument as length
		vec_len = atoi( argv[1] );
	printf( "Add two vectors of length %d\n", vec_len );

	// fill two vectors with random numbers
	size_t size = vec_len * sizeof(float);
	float * h_a = (float*) malloc( size );
	float * h_b = (float*) malloc( size );
	float * h_c = (float*) malloc( size );
	if ( h_a == NULL || h_b == NULL || h_c == NULL )
    {
        fprintf(stderr, "Failed to allocate host vectors!\n");
		exit(1);
	}

	srand48( unsigned(time(NULL)) );
    for ( unsigned int i = 0; i < vec_len; i++ )
    {
        h_a[i] = drand48();
        h_b[i] = drand48();
    }

	// transfer the arrays to the device
	float *d_a, *d_b, *d_c;
	cudaMalloc( (void **) &d_a, size );
	cudaMalloc( (void **) &d_b, size );
	cudaMalloc( (void **) &d_c, size );
	handleCUDAerror(__LINE__);

	cudaMemcpy( d_a, h_a, size, cudaMemcpyHostToDevice );
	cudaMemcpy( d_b, h_b, size, cudaMemcpyHostToDevice );
	handleCUDAerror(__LINE__);									// yes, we should do this after every memcpy ...

	// set layout of blocks and threads; we make a 1D threads/blocks layout here
	dim3 threads(16,16);										// = 256 threads per block
	int n_threads_pb = threads.x * threads.y;
	int n_blocks =  (vec_len + n_threads_pb - 1) / n_threads_pb;
	int nb_sqrt = (int)( ceilf( sqrtf( n_blocks ) ) );			// integer square root of n_blocks, rounded up
	dim3 blocks( nb_sqrt, nb_sqrt );

	printf( "Threads per block = %d\n", n_threads_pb );
	printf( "Blocks per grid   = %d\n", nb_sqrt );

	checkMaxProperties( blocks, threads );

	// launch the vector addition kernel
	addVectors<<<blocks,threads>>>( d_a, d_b, d_c, vec_len );
	handleCUDAerror(__LINE__);

	// copy results back to the host (does cudaDeviceSynchronize implicitely)
	cudaMemcpy( h_c, d_c, size, cudaMemcpyDeviceToHost );
	handleCUDAerror(__LINE__);

	// verify the results (kind of a unit test)
	bool verify_ok = true;
	for ( unsigned int i = 0; i < vec_len; i++ )
	{
		if ( fabsf(h_a[i] + h_b[i] - h_c[i]) > 1e-5 )
		{
			fprintf(stderr, "Error at element %d: %f + %f != %f !\n",
					i, h_a[i], h_b[i], h_c[i] );
			verify_ok = false;
		}
	}

	// free all memory
	cudaFree( d_a );
	cudaFree( d_b );
	cudaFree( d_c );
	handleCUDAerror(__LINE__);

	free( h_a );
	free( h_b );
	free( h_c );

	cudaDeviceReset();
	if ( verify_ok )
		puts("Success.");
	return 0;
}

